#===============================================================================
# bbs.py
#===============================================================================

# Imports ======================================================================

from functools import partial
from multiprocessing import Pool
from betabernsum.region import region
from betabernsum.probability_mass import (
    probability_mass_independent, probability_mass_dependent
)




# Functions ====================================================================

def bbs_pmf(k, n, a, b, independent=True):
    """Probability mass function for a sum of beta-bernoulli variables

    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group
    independent : bool
        If TRUE (default), assume a sum of two independent groups of variables.
        If FALSE, assume all variables are mutually dependent.

    Returns
    -------
    float
        the value of the PMF
    """

    if k > sum(n) / 2:
        k, a, b = sum(n) - k, b, a

    if isinstance(n, int) or (isinstance(n, (tuple, list)) and len(n) == 1):
        reg = ((k,),)
    else:
        reg = region(k, n)
    return sum(
        probability_mass_independent(coord, n, a, b) if independent
        else probability_mass_dependent(coord, n, a, b)
        for coord in reg
    )

def bbs_cdf(
    k,
    n,
    a,
    b,
    independent=True,
    processes=1,
    max_iter=float('inf'),
    graceful=False
):
    """Cumulative distribution function for a sum of beta-bernoulli variables

    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group
    independent : bool
        If TRUE (default), assume a sum of two independent groups of variables.
        If FALSE, assume all variables are mutually dependent.
    processes : int
        number of processes to use
    max_iter : int
        maximum number of iterations
    graceful
        if True, return None when max_iter is exceeded, rather than raising an
        error

    Returns
    -------
    float
        the value of the CDF
    """

    if k == sum(n):
        return 1.0

    if k <= sum(n) / 2:
        speed_flip, r = False, range(k + 1)
    else:
        speed_flip, r, a, b = True, range(sum(n) - k), b, a

    if len(r) > max_iter:
        if graceful:
            return None
        else:
            raise RuntimeError('max iter exceeded')

    if processes == 1:
        return speed_flip + (1 - 2 * speed_flip) * sum(
            map(partial(bbs_pmf, n=n, a=a, b=b, independent=independent), r)
        )
    else:
        with Pool(processes=min(len(r), processes)) as pool:
            return speed_flip + (1 - 2 * speed_flip) * sum(
                pool.map(
                    partial(bbs_pmf, n=n, a=b, b=a, independent=independent), r
                )
            )



#===============================================================================
# probability_mass.py
#===============================================================================

# Imports ======================================================================

from functools import reduce
from math import exp, log
from operator import mul
from scipy.integrate import quad
from scipy.misc import comb
from scipy.special import beta as beta_function
from scipy.stats import beta




# Functions ====================================================================

def prod(iterable):
    """"Product of values in an iterable
    
    Parameters
    ----------
    iterable
        an iterable
    
    Returns
    -------
    the product
    """

    return reduce(mul, iterable, 1)


def betabinom_pmf(k, n, a, b):
    """Probability mass function for a beta binomial distribution

    Parameters
    ----------
    k : int
        number of successes
    n : int
        number of trials
    a : float
        first (positive) shape parameter
    b : float
        second (positive) shape parameter

    Returns
    -------
    float
        the probability mass
    """

    return comb(n, k) * beta_function(k + a, n - k + b) / beta_function(a, b)


def probability_mass_independent(k, n, a, b):
    """Probability mass for a bbs in the independent case

    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group
    
    Returns
    -------
    float
        the probability mass
    """
    
    assert len(k) == len(n) == len(a) == len(b)
    return prod(
        betabinom_pmf(k_i, n_i, a_i, b_i)
        for k_i, n_i, a_i, b_i in zip(k, n, a, b)
    )


def evaluate_factor(t, k, n, a, b):
    """Evaluate a factor for the product in `integrand_dependent`

    t : float
        a value on the domain of integration
    k : int
        a number of successes
    n : int
        a number of trials
    a : float
        a first shape parameter
    b : float
        a second shape parameter
    
    Returns
    -------
    float
        value of a factor
    """

    q = beta.ppf(t, a, b)
    if k == 0:
        return (1 - q)**n
    elif k == n:
        return q**k
    else:
        return q**k * (1 - q)**(n - k)
    

def integrand_dependent(t, k, n, a, b):
    """The integrand for computing probability mass in the dependent case

    Parameters
    ----------
    t
        value from the domain of integration
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group

    Returns
    -------
    tuple
        the values of the integrand
    """

    assert len(k) == len(n) == len(a) == len(b)
    return prod(
        evaluate_factor(t, k_j, n_j, a_j, b_j)
        for k_j, n_j, a_j, b_j in zip(k, n, a, b)
    )


def log_integral_dependent(k, n, a, b):
    """Get logarithm of integral
    
    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group

    Returns
    -------
    float
        the value of the log integral
    """
    
    return log(quad(integrand_dependent, 0, 1, args = (k, n, a, b))[0])


def log_coefficient_dependent(k, n):
    """get logarithm of coefficient

    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    
    Returns
    -------
    float
        the value of the log coefficient
    """

    return sum(log(comb(n_i, k_i)) for k_i, n_i in zip(k, n))


def probability_mass_dependent(k, n, a, b):
    """Probability mass for a bbs in the dependent case

    Parameters
    ----------
    k
        iterable giving the number of successes for each group
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group
    
    Returns
    -------
    float
        the probability mass
    """

    return exp(
        log_coefficient_dependent(k, n)
        + log_integral_dependent(k, n, a, b)
    )



#===============================================================================
# bbs_test.py
#===============================================================================

# Imports ======================================================================

from betabernsum.bbs import bbs_pmf, bbs_cdf




# Functions ====================================================================

def bbs_test(
    x,
    n,
    a,
    b,
    independent=True,
    alternative='two-sided',
    processes=1,
    max_iter=float('inf'),
    graceful=False
):
    """Perform a hypothesis test using a BBS distribution
    
    Parameters
    ----------
    x : int
        the number of successes
    n
        iterable giving the number of trials for each group
    a
        iterable giving the first shape parameter for each group
    b
        iterable giving the second shape parameter for each group
    alternative : str
        alternative hypothesis {'less', 'greater', 'two-sided'}
    processes : int
        number of processes to use
    max_iter : int
        maximum number of iterations
    graceful
        if True, return None when max_iter is exceeded, rather than raising an
        error

    Returns
    -------
    float
        the p-value of the hypothesis test
    """
    
    lower_tail_area = bbs_cdf(
        x, n, a, b, independent=independent, processes=processes,
        max_iter=max_iter, graceful=graceful
    )
    if lower_tail_area is None:
        return None
    if alternative == 'less':
        return lower_tail_area
    else:
        upper_tail_area = 1 - lower_tail_area + bbs_pmf(
            x, n, a, b, independent=independent
        )
    if alternative == 'two-sided':
        return min(1, 2 * min(lower_tail_area, upper_tail_area))
    elif alternative == 'greater':
        return upper_tail_area



"""Compute sums of beta-bernoulli variables

Functions
---------
bbs_pmf
    Probability mass function for BBS
bbs_cdf
    Cumulative distribution function for BBS
bbs_test
    Hypothesis test for BBS
"""

from betabernsum.bbs import bbs_pmf, bbs_cdf
from betabernsum.bbs_test import bbs_test